from . import ops
from .ops import *
from .autograd import Tensor, cpu, all_devices

from . import init
from .init import ones, zeros, zeros_like, ones_like, randn

from . import nn
from . import optim
from . import data


from typing import Optional
from ..autograd import Op, Tensor
from .ops_mathematic import *
import numpy as array_api


class LogSumExp(Op):
    def __init__(self, axes: Optional[tuple] = None):
        super().__init__()
        self.axes = axes

    def compute(self, Z: Tensor):
        max_z = array_api.max(Z, axis=self.axes, keepdims=True)
        max_z_reduce = array_api.max(Z, axis=self.axes)
        return array_api.log(array_api.sum(array_api.exp(Z - max_z), axis=self.axes)) + max_z_reduce

    def gradient(self, out_grad, node):
        z = node.inputs[0]
        max_z = z.underly().max(self.axes, keepdims=True)
        exp_z = exp(z - max_z)
        sum_exp_z = sum(exp_z, self.axes)
        grad_sum_exp_z = out_grad / sum_exp_z
        expand_shape = list(z.shape)
        axes = range(len(expand_shape)) if self.axes is None else self.axes
        if isinstance(axes, int):
            axes = [axes]
        for axis in axes:
            expand_shape[axis] = 1
        grad_exp_z = grad_sum_exp_z.reshape(expand_shape).broadcast_to(z.shape)
        return grad_exp_z * exp_z


def logsumexp(input, axes=None):
    return LogSumExp(axes=axes)(input)

class LogSoftmax(Op):
    def __init__(self, axes: Optional[tuple] = None):
        self.axes = axes
    
    def compute(self, Z: Tensor) -> Tensor:
        pass
        

def logsoftmax(input: Tensor, axes=None):
    return LogSoftmax(axes=axes)(input)


import flowlite as fl
import flowlite.nn as nn


class Optimizer:
    def __init__(self, params: nn.Parameter):
        self.params = params
    
    def step(self):
        raise NotImplementedError()
    
    def zero_grad(self):
        for p in self.params:
            p.grad = None

class SGD(Optimizer):
    def __init__(self, params: nn.Parameter, lr=0.01, momentum=0.0, weight_decay=0.0):
        super().__init__(params)
        self.lr = lr
        self.momentum = momentum
        self.u = {}
        self.weight_decay = weight_decay

    def step(self):
        '''
        u_{t+1} = \beta u_t + (1 - \beta) \nabla_{\theta}f(\theta_t)
        \theta_{t+1} = \theta_t - \alpha u_{t+1}
        
        \alpha: lr
        \beta: momentum
        '''
        # https://discuss.pytorch.org/t/how-does-sgd-weight-decay-work/33105/2
        for param in self.params:
            grad = param.grad.data + self.weight_decay * param.data
            grad = self.u.get(param, 0) * self.momentum + (1 - self.momentum) * grad
            #TODO: figure out if it should be ndl.Tensor(grad, dtype=param.dtype, requires_grad=False)
            self.u[param] = fl.Tensor(grad, dtype=param.dtype)
            param.data -= self.lr * self.u[param]


class Adam(Optimizer):
    def __init__(
        self,
        params,
        lr=0.01,
        beta1=0.9,
        beta2=0.999,
        eps=1e-8,
        weight_decay=0.0,
    ):
        super().__init__(params)
        self.lr = lr
        self.beta1 = beta1
        self.beta2 = beta2
        self.eps = eps
        self.weight_decay = weight_decay
        self.t = 0

        self.m = {}
        self.v = {}

    def step(self):
        # self.t += 1
        # for w in self.params:
        #     grad = w.grad.data + self.weight_decay * w.data
        #     self.m[w] = self.beta1 * self.m.get(w, 0) + (1 - self.beta1) * grad
        #     self.v[w] = self.beta2 * self.v.get(w, 0) + (1 - self.beta2) * (grad ** 2)
        #     unbiased_m = self.m[w] / (1 - self.beta1 ** self.t)
        #     unbiased_v = self.v[w] / (1 - self.beta2 ** self.t)
        #     w.data = w.data - self.lr * unbiased_m / (unbiased_v**0.5 + self.eps)
        self.t += 1
        for param in self.params:
            grad = param.grad.data + self.weight_decay * param.data
            self.m[param] = self.beta1 * self.m.get(param, 0) + (1 - self.beta1) * grad
            self.v[param] = self.beta2 * self.v.get(param, 0) + (1 - self.beta2) * (grad ** 2)
            m_hat = (self.m[param] / (1 - self.beta1 ** self.t)).detach()
            v_hat = (self.v[param] / (1 - self.beta2 ** self.t)).detach()
            update = fl.Tensor(self.lr * m_hat / (v_hat ** 0.5 + self.eps), dtype=param.dtype).detach()
            param.data -= update.detach()

#TODO: more optim, like AdaGrad, RMSprop


import flowlite
from .backend_numpy import Device, cpu, all_devices
from typing import List, Optional, Dict, Tuple, Union
import numpy

from flowlite import init

TENSOR_COUNTER = 0

import numpy as array_api

NDArray = numpy.ndarray


class Op:
    def __call__(self, *args):
        return Tensor.make_from_op(self, args)

    def compute(self, *args: Tuple[NDArray]) -> NDArray:
        raise NotImplementedError()
    
    def gradient(self, out_grad: 'Tensor', node: 'Tensor') -> Union['Tensor', Tuple['Tensor']]:
        raise NotImplementedError()

    def gradient_as_tuple(self, out_grad: 'Tensor', node: 'Tensor') -> Tuple['Tensor']:
        """Convenience method to always return a tuple from gradient call"""
        output = self.gradient(out_grad, node)
        if isinstance(output, tuple):
            return output
        elif isinstance(output, list):
            return tuple(output)
        else:
            return (output,)

class Tensor:
    op: Optional[Op]
    inputs: List['Tensor']
    underlying_data: NDArray
    requires_grad: bool
    grad: Optional['Tensor']
    
    def __init__(
        self,
        array: Union['Tensor', NDArray],
        *,
        device: Optional[Device] = None,
        dtype = None,
        requires_grad: bool = False,
        **kwargs
    ) -> None:
        if isinstance(array, Tensor):
            device = array.device if device is None else device
            dtype = array.dtype if dtype is None else dtype
            if device == array.device and dtype == array.dtype:
                underlying_data = array.underlying_data
            # TODO: _array_from_numpy should be in NDArray
            else:
                underlying_data = Tensor._array_from_numpy(
                    array.numpy(), device=device, dtype=dtype
                )
        
        else:
            device = device if device else cpu()
            underlying_data = Tensor._array_from_numpy(array, device=device, dtype=dtype)
        
        self._init(None, [], underlying_data=underlying_data, requires_grad=requires_grad)
    
    def _init(self, 
              op: Optional[Op], 
              inputs: List['Tensor'], 
              *, 
              num_outputs: int = 1, 
              underlying_data: Optional[NDArray] = None, 
              requires_grad: Optional[bool] = None
        ) -> None:
        global TENSOR_COUNTER
        TENSOR_COUNTER += 1
        if requires_grad is None:
            requires_grad = any(x.requires_grad for x in inputs)
        self.op = op
        self.inputs = inputs
        self.num_outputs = num_outputs
        self.underlying_data = underlying_data
        self.requires_grad = requires_grad
        self.grad = None
    
    def underly(self) -> NDArray:
        return self.underlying_data
        
    
    @staticmethod
    def _array_from_numpy(numpy_array: numpy.ndarray, device, dtype) -> NDArray:
        if array_api is numpy:
            return numpy.array(numpy_array, dtype=dtype)
        return array_api.array(numpy_array, device=device, dtype=dtype)
    
    @staticmethod
    def make_from_op(op: Op, inputs: List['Tensor']) -> 'Tensor':
        tensor = Tensor.__new__(Tensor)
        tensor._init(op, inputs)
        
        tensor.underlying_data = tensor.op.compute(
            *[x.underly() for x in tensor.inputs]
        )
        
        if not tensor.requires_grad:
            return tensor.detach()
        
        return tensor
    
    @staticmethod
    def make_const(data: NDArray, requires_grad: bool = False) -> 'Tensor':
        """Create a new tensor that shares the data but detaches from the graph."""
        '''Used in detach()'''
        tensor = Tensor.__new__(Tensor)
        tensor._init(
            None,
            [],
            underlying_data=data,
            requires_grad=requires_grad,
        )
        return tensor

    @property
    def data(self) -> 'Tensor':
        '''
        To be consistent with pytorch, this method is used to 'detach' a data,
        not to get the underlying data
        the requires_grad would be set to False
        '''
        return self.detach()
    
    @data.setter
    def data(self, value):
        assert isinstance(value, Tensor)
        assert value.dtype == self.dtype, f'{value.dtype} {self.dtype}'        
        self.underlying_data = value.underly() 

    def detach(self) -> 'Tensor':
        """Create a new tensor that shares the data but detaches from the graph."""
        return Tensor.make_const(self.underly())

    @property
    def shape(self):
        return self.underlying_data.shape
    
    @property
    def dtype(self):
        return self.underlying_data.dtype
    
    @property
    def device(self):
        if array_api is numpy:
            return cpu()
        return self.underlying_data.device


    def backward(self, out_grad: Optional['Tensor'] = None):
        out_grad = (
            out_grad
            if out_grad
            else init.ones(*self.shape, dtype=self.dtype, device=self.device)
        )
        compute_gradient_of_variables(self, out_grad)

    def is_leaf(self) -> bool:
        return self.op is None
    
    def __del__(self):
        global TENSOR_COUNTER
        TENSOR_COUNTER -= 1
    
    def __repr__(self) -> str:
        s = f'fl.Tensor(\n{str(self.underlying_data)})'
        if self.requires_grad:
            s += ', requires_grad=True'
        return s
    
    def __str__(self) -> str:
        return self.__repr__()
    
    def numpy(self):
        if array_api is numpy:
            return self.underlying_data
        return self.underlying_data.numpy()
    

    def __add__(self, other):
        if isinstance(other, Tensor):
            return flowlite.ops.EWiseAdd()(self, other)
        else:
            return flowlite.ops.AddScalar(other)(self)

    def __mul__(self, other):
        if isinstance(other, Tensor):
            return flowlite.ops.EWiseMul()(self, other)
        else:
            return flowlite.ops.MulScalar(other)(self)

    def __pow__(self, other):
        if isinstance(other, Tensor):
            return flowlite.ops.EWisePow()(self, other)
        else:
            return flowlite.ops.PowerScalar(other)(self)

    def __sub__(self, other):
        if isinstance(other, Tensor):
            return flowlite.ops.EWiseAdd()(self, flowlite.ops.Negate()(other))
        else:
            return flowlite.ops.AddScalar(-other)(self)

    def __rsub__(self, other):
        if isinstance(other, Tensor):
            return flowlite.ops.EWiseAdd()(flowlite.ops.Negate()(self), other)
        else:
            return flowlite.ops.AddScalar(other)(-self)

    def __truediv__(self, other):
        if isinstance(other, Tensor):
            return flowlite.ops.EWiseDiv()(self, other)
        else:
            return flowlite.ops.DivScalar(other)(self)

    def __matmul__(self, other):
        return flowlite.ops.MatMul()(self, other)

    def matmul(self, other):
        return flowlite.ops.MatMul()(self, other)

    def sum(self, dim = None, keepdim: bool = False):
        return flowlite.ops.Sum(dim, keepdim)(self)

    def broadcast_to(self, shape):
        return flowlite.ops.BroadcastTo(shape)(self)

    def reshape(self, shape):
        return flowlite.ops.Reshape(shape)(self)

    def __neg__(self):
        return flowlite.ops.Negate()(self)

    def transpose(self, dim0: int, dim1: int):
        return flowlite.ops.Transpose(dim0=dim0, dim1=dim1)(self)
    
    #TODO: add tensor.T(), checkout args in pytorch

    __radd__ = __add__
    __rmul__ = __mul__
    __rmatmul__ = __matmul__
        
        
def compute_gradient_of_variables(output_tensor: Tensor, out_grad: Tensor):
    """Take gradient of output node with respect to each node in node_list.

    Store the computed result in the grad field of each Variable.
    """
    # a map from node to a list of gradient contributions from each output node
    node_to_output_grads_list: Dict[Tensor, List[Tensor]] = {}
    # Special note on initializing gradient of
    # We are really taking a derivative of the scalar reduce_sum(output_node)
    # instead of the vector output_node. But this is the common case for loss function.
    node_to_output_grads_list[output_tensor] = [out_grad]

    # Traverse graph in reverse topological order given the output_node that we are taking gradient wrt.
    reverse_topo_order = list(reversed(find_topo_sort([output_tensor])))

    for node in reverse_topo_order:
        # sum up partial ajoints
        ajoint = sum_node_list(node_to_output_grads_list[node])
        node.grad = ajoint
        if node.op is None:
            # Leaf node
            continue
        # compute partial ajoints for input node
        partial_ajoints = node.op.gradient_as_tuple(ajoint, node)
        for in_node, partial_ajoint in zip(node.inputs, partial_ajoints):
            if in_node not in node_to_output_grads_list:
                node_to_output_grads_list[in_node] = []
            node_to_output_grads_list[in_node].append(partial_ajoint)



def find_topo_sort(node_list: List[Tensor]) -> List[Tensor]:
    """Given a list of nodes, return a topological sort list of nodes ending in them.

    A simple algorithm is to do a post-order DFS traversal on the given nodes,
    going backwards based on input edges. Since a node is added to the ordering
    after all its predecessors are traversed due to post-order DFS, we get a topological
    sort.
    """
    visited = set()
    topo_order = []
    for node in node_list:
        if node not in visited:
            topo_sort_dfs(node, visited, topo_order)
    return topo_order


def topo_sort_dfs(node, visited, topo_order):
    """Post-order DFS"""
    for input_node in node.inputs:
        if input_node not in visited:
            topo_sort_dfs(input_node, visited, topo_order)
    visited.add(node)
    topo_order.append(node)



def sum_node_list(node_list: List[Tensor]):
    """Custom sum function in order to avoid create redundant nodes in Python sum implementation."""
    from operator import add
    from functools import reduce

    return reduce(add, node_list)



from typing import Optional, Union, Tuple

from ..autograd import NDArray
from ..autograd import Op, Tensor

import numpy as array_api
#TODO: for all ops, maybe super().__init()__ in __init__
class EWiseAdd(Op):
    def compute(self, a: NDArray, b: NDArray) -> NDArray:
        return a + b

    def gradient(self, out_grad: Tensor, node: Tensor) -> Tensor:
        return out_grad, out_grad
    
def add(a, b):
    return EWiseAdd()(a, b)


class AddScalar(Op):
    def __init__(self, scalar):
        self.scalar = scalar

    def compute(self, a: NDArray) -> NDArray:
        return a + self.scalar

    def gradient(self, out_grad: Tensor, node: Tensor) -> Tensor:
        return out_grad


def add_scalar(a, scalar):
    return AddScalar(scalar)(a)


class EWiseMul(Op):
    def compute(self, a: NDArray, b: NDArray) -> NDArray:
        return a * b

    def gradient(self, out_grad: Tensor, node: Tensor) -> Tensor:
        lhs, rhs = node.inputs
        return out_grad * rhs, out_grad * lhs


def multiply(a, b):
    return EWiseMul()(a, b)


class MulScalar(Op):
    def __init__(self, scalar):
        self.scalar = scalar

    def compute(self, a: NDArray) -> NDArray:
        return a * self.scalar

    def gradient(self, out_grad: Tensor, node: Tensor) -> Tensor:
        return (out_grad * self.scalar,)


def mul_scalar(a, scalar):
    return MulScalar(scalar)(a)


class PowerScalar(Op):
    def __init__(self, scalar: int):
        self.scalar = scalar

    def compute(self, a: NDArray) -> NDArray:
        return a ** self.scalar

    def gradient(self, out_grad: Tensor, node: Tensor) -> Tensor:
        return out_grad * self.scalar * node.inputs[0] ** (self.scalar - 1)         


def power_scalar(a, scalar):
    return PowerScalar(scalar)(a)


class EWisePow(Op):
    def compute(self, a: NDArray, b: NDArray) -> NDArray:
        return a**b

    def gradient(self, out_grad: Tensor, node: Tensor) -> Tensor:
        if not isinstance(node.inputs[0], NDArray) or not isinstance(
            node.inputs[1], NDArray
        ):
            raise ValueError("Both inputs must be tensors (NDArray).")

        a, b = node.inputs[0], node.inputs[1]
        grad_a = out_grad * b * (a ** (b - 1))
        grad_b = out_grad * (a**b) * array_api.log(a.data)
        return grad_a, grad_b

def power(a, b):
    return EWisePow()(a, b)


class EWiseDiv(Op):
    def compute(self, a: NDArray, b: NDArray):
        return a / b

    def gradient(self, out_grad: Tensor, node: Tensor):
        a, b = node.inputs
        grad_a = out_grad / b
        grad_b = out_grad * (-a / (b**2))
        return grad_a, grad_b


def divide(a, b):
    return EWiseDiv()(a, b)


class DivScalar(Op):
    def __init__(self, scalar):
        self.scalar = scalar

    def compute(self, a):
        return a / self.scalar

    def gradient(self, out_grad: Tensor, node: Tensor):
        return out_grad / self.scalar


def divide_scalar(a, scalar):
    return DivScalar(scalar)(a)


class Transpose(Op):
    def __init__(self, dim0: int, dim1: int):
        self.dim0 = dim0
        self.dim1 = dim1

    def compute(self, x: NDArray) -> NDArray:
        return array_api.swapaxes(x, self.dim0, self.dim1)

    def gradient(self, out_grad, node):
        return out_grad.transpose(self.dim0, self.dim1)


def transpose(x: Tensor, dim0: int, dim1: int) -> Tensor:
    return Transpose(dim0=dim0, dim1=dim1)(x)


# TODO: to be consistent with pytorch
class Reshape(Op):
    def __init__(self, shape):
        self.shape = shape

    def compute(self, x: NDArray) -> NDArray:
        return array_api.reshape(x, self.shape)

    def gradient(self, out_grad, node):
        return out_grad.reshape(node.inputs[0].shape)


def reshape(a, shape):
    return Reshape(shape)(a)


class BroadcastTo(Op):
    def __init__(self, shape):
        self.shape = shape

    def compute(self, a):
        return array_api.broadcast_to(a, self.shape)

    def gradient(self, out_grad, node):
        ori_shape = node.inputs[0].shape
        shrink_dims = [i for i in range(len(self.shape))]
        for i, (ori, cur) in enumerate(zip(reversed(ori_shape), reversed(self.shape))):
            if ori == cur:
                shrink_dims[len(self.shape) - i - 1] = -1
        shrink_dims = tuple(filter(lambda x: x >= 0, shrink_dims))
        return out_grad.sum(shrink_dims).reshape(ori_shape)


def broadcast_to(a, shape):
    return BroadcastTo(shape)(a)


class Sum(Op):
    def __init__(self, dim: Union[int, Tuple[int, ...]] = None, keepdim: bool = False):
        self.dim = dim
        self.keepdim = keepdim

    def compute(self, a):
        return array_api.sum(a, axis=self.dim, keepdims=self.keepdim)

    def gradient(self, out_grad: Tensor, node: Tensor) -> Tensor:
        #TODO: not sure for keepdim == True
        if self.keepdim == True:
            return out_grad.broadcast_to(node.inputs[0].shape)

        new_shape = list(node.inputs[0].shape)
        dims = range(len(new_shape)) if self.dim is None else self.dim
        if isinstance(self.dim, int):
            dims = [self.dim]  # Convert single integer to a list
        for dim in dims:
            new_shape[dim] = 1
        return out_grad.reshape(new_shape).broadcast_to(node.inputs[0].shape)


def sum(x: Tensor, dim = None, keepdim: bool = False):
    return Sum(dim, keepdim)(x)


class MatMul(Op):
    def compute(self, a: NDArray, b: NDArray) -> NDArray:
        return array_api.matmul(a, b)

    def gradient(self, out_grad, node):
        lhs, rhs = node.inputs
        lhs_ndim = len(lhs.shape)
        rhs_ndim = len(rhs.shape)
        #TODO: maybe let transpose accept None
        lgrad, rgrad = matmul(out_grad, rhs.transpose(rhs_ndim - 1, rhs_ndim - 2)), matmul(lhs.transpose(lhs_ndim - 1, lhs_ndim - 2), out_grad)
        
        if len(lhs.shape) < len(lgrad.shape):
            lgrad = lgrad.sum(tuple([i for i in range(len(lgrad.shape) - len(lhs.shape))]))
        if len(rhs.shape) < len(rgrad.shape):
            rgrad = rgrad.sum(tuple([i for i in range(len(rgrad.shape) - len(rhs.shape))]))
        return lgrad, rgrad


def matmul(a, b):
    return MatMul()(a, b)


class Negate(Op):
    def compute(self, a):
        return -a

    def gradient(self, out_grad, node):
        return -out_grad


def negate(a):
    return Negate()(a)


class Log(Op):
    def compute(self, a):
        return array_api.log(a)

    def gradient(self, out_grad, node):
        return out_grad / node.inputs[0]


def log(a):
    return Log()(a)


class Exp(Op):
    def compute(self, a):
        return array_api.exp(a)

    def gradient(self, out_grad, node):
        return out_grad * exp(node.inputs[0])


def exp(a):
    return Exp()(a)


#TODO: add inplace=True
class ReLU(Op):
    # def __init__(inplace: bool = False):

    def compute(self, a):
        out = a.copy()
        out[out < 0] = 0
        return out

    def gradient(self, out_grad, node):
        out = node.underly().copy()
        out[out > 0] = 1
        out[out <= 0] = 0
        return out_grad * Tensor(out)


def relu(a):
    return ReLU()(a)



from typing import List, Optional
from ..data_basic import Dataset
import numpy as np
import gzip, struct

class MNISTDataset(Dataset):
    def __init__(
        self,
        image_filename: str,
        label_filename: str,
        transforms: Optional[List] = None,
    ):
        with gzip.open(label_filename, "rb") as label:
            magic, n = struct.unpack('>2I', label.read(8))
            y = np.frombuffer(label.read(), dtype=np.uint8)

        with gzip.open(image_filename, "rb") as image:
            magic, num, rows, cols = struct.unpack('>4I', image.read(16))
            X = np.frombuffer(image.read(), dtype=np.uint8).reshape(len(y), 784)
    
        X = (X.astype(np.float32) / 255).reshape((-1, 28, 28, 1))
        self.images = X
        self.labels = y
        self.transforms = transforms

    def __getitem__(self, index) -> object:
        img = self.apply_transforms(self.images[index])
        return img, self.labels[index]

    def __len__(self) -> int:
        return self.images.shape[0]


"""This file defies specific implementations of devices when using numpy as NDArray backend.
"""
import numpy


class Device:
    """Baseclass of all device"""


class CPUDevice(Device):
    """Represents data that sits in CPU"""

    def __repr__(self):
        return "pytensor.cpu()"

    def __hash__(self):
        return self.__repr__().__hash__()

    def __eq__(self, other):
        return isinstance(other, CPUDevice)

    def enabled(self):
        return True

    def zeros(self, *shape, dtype="float32"):
        return numpy.zeros(shape, dtype=dtype)

    def ones(self, *shape, dtype="float32"):
        return numpy.ones(shape, dtype=dtype)

    def randn(self, *shape):
        # note: numpy doesn't support types within standard random routines, and
        # .astype("float32") does work if we're generating a singleton
        return numpy.random.randn(*shape)

    def rand(self, *shape):
        # note: numpy doesn't support types within standard random routines, and
        # .astype("float32") does work if we're generating a singleton
        return numpy.random.rand(*shape)

    def one_hot(self, n, i, dtype="float32"):
        return numpy.eye(n, dtype=dtype)[i]

    def empty(self, shape, dtype="float32"):
        return numpy.empty(shape, dtype=dtype)

    def full(self, shape, fill_value, dtype="float32"):
        return numpy.full(shape, fill_value, dtype=dtype)


def cpu():
    """Return cpu device"""
    return CPUDevice()


def default_device():
    return cpu()


def all_devices():
    """return a list of all available devices"""
    return [cpu()]


import math
from .init_basic import *

def xavier_uniform(fan_in: int, fan_out: int, gain: float = 1.0, **kwargs):
    a = gain * math.sqrt(6 / (fan_in + fan_out))
    return rand(fan_in, fan_out, low=-a, high=a, **kwargs)

def xavier_normal(fan_in: int, fan_out: int, gain: float = 1.0, **kwargs):
    std = gain * math.sqrt(2 / (fan_in + fan_out))
    return randn(fan_in, fan_out, std=std, **kwargs)   

def kaiming_uniform(fan_in: int, fan_out: int, nonlinearity='relu', **kwargs):
    assert nonlinearity == 'relu', 'Only relu supported currently'
    bound = math.sqrt(6 / fan_in)
    return rand(fan_in, fan_out, low=-bound, high=bound, **kwargs)


def kaiming_normal(fan_in, fan_out, nonlinearity='relu', **kwargs):
    assert nonlinearity == 'relu', 'Only relu supported currently'
    bound = math.sqrt(2 / fan_in)
    return randn(fan_in, fan_out, std=bound, **kwargs)


from typing import List
from flowlite.autograd import Tensor
from flowlite import ops
import flowlite.init as init


class Parameter(Tensor):
    """A special kind of tensor that represents parameters."""
    def __init__(self, x: Tensor):
        super().__init__(x, requires_grad=True)


def _unpack_params(value: object) -> List[Tensor]:
    if isinstance(value, Parameter):
        return [value]
    elif isinstance(value, Module):
        return value.parameters()
    elif isinstance(value, dict):
        params = []
        for k, v in value.items():
            params += _unpack_params(v)
        return params
    elif isinstance(value, (list, tuple)):
        params = []
        for v in value:
            params += _unpack_params(v)
        return params
    else:
        return []

def _child_modules(value: object) -> List['Module']:
    if isinstance(value, Module):
        modules = [value]
        modules.extend(_child_modules(value.__dict__))
        return modules
    if isinstance(value, dict):
        modules = []
        for k, v in value.items():
            modules += _child_modules(v)
        return modules
    elif isinstance(value, (list, tuple)):
        modules = []
        for v in value:
            modules += _child_modules(v)
        return modules
    else:
        return []

class Module:
    def __init__(self):
        self.training = True
    
    def parameters(self) -> List[Tensor]:
        return _unpack_params(self.__dict__)

    def _children(self) -> List['Module']:
        return _child_modules(self.__dict__)

    def eval(self):
        self.training = False
        for m in self._children():
            m.training = False
    
    def train(self):
        self.training = True
        for m in self._children():
            m.training = True

    def __call__(self, *args, **kwargs):
        return self.forward(*args, **kwargs)


class Identity(Module):
    def forward(self, x: Tensor) -> Tensor:
        return x

class Linear(Module):
    def __init__(self, in_features: int, out_features: int, bias: bool = True, device=None, dtype=None):
        super().__init__()
        self.weight = Parameter(init.kaiming_uniform(in_features, out_features))
        if bias:
            self.bias = Parameter(init.kaiming_uniform(out_features, 1).transpose(0, 1))
        else:
            bias = None
    
    def forward(self, x: Tensor) -> Tensor:
        out = x @ self.weight
        if self.bias:
            out += self.bias.broadcast_to(out.shape)
        return out

class ReLU(Module):
    # TODO: about inplace=True
    def forward(self, x: Tensor) -> Tensor:
        return ops.relu(x)

class Sequential(Module):
    def __init__(self, *modules):
        super().__init__()
        self.modules = modules

    def forward(self, x: Tensor) -> Tensor:
        from functools import reduce
        return reduce(lambda out, module: module(out), self.modules, x)

class Flatten(Module):
    def forward(self, x: Tensor) -> Tensor:
        from functools import reduce
        size = reduce(lambda a, b: a * b, x.shape)
        # we assume the first dim is batch_size
        return x.reshape((x.shape[0], size // x.shape[0]))


class CrossEntropyLoss(Module):
    def forward(self, logits: Tensor, y: Tensor) -> Tensor:
        '''
        logits: (bs, num_classes), 
        '''
        one_hot_y = init.one_hot(logits.shape[1], y)
        return ops.sum(ops.logsumexp(logits, axes=1) - (logits * one_hot_y).sum(dim=1)) / logits.shape[0]


# TODO: inplace
class Dropout(Module):
    def __init__(self, p: float = 0.5, inplace: bool = False):
        super().__init__()
        self.p = p
        self.inplace = inplace

    def forward(self, x: Tensor) -> Tensor:
        if self.training:
            mask = init.randb(*x.shape, p=1-self.p)
            return x * mask / (1 - self.p)
        else:
            return x

#TODO: all the device and dtype parameters in bn and ln
class BatchNorm1d(Module):
    def __init__(self, num_features: int, eps=1e-5, momentum=0.1, device=None, dtype=None):
        super().__init__()
        self.eps = eps
        self.momentum = momentum
        
        self.weight = Parameter(init.ones(num_features))
        self.bias = Parameter(init.zeros(num_features))
        self.running_mean = init.zeros(num_features)
        self.running_var = init.ones(num_features)
    
    def forward(self, x: Tensor) -> Tensor:
        # x: (batch_size, C) C is #feature or #channels
        if self.training:
            mean = x.sum(dim=0, keepdim=True).broadcast_to(x.shape) / x.shape[0]
            var = ((x - mean)**2).sum(dim=0, keepdim=True).broadcast_to(x.shape) / x.shape[0]
            #TODO: not to * mean.underly()[0]
            self.running_mean = (1 - self.momentum) * self.running_mean + self.momentum * mean.underly()[0]
            self.running_var = (1 - self.momentum) * self.running_var + self.momentum * var.underly()[0]
            norm = (x - mean) / ((var + self.eps) ** 0.5)
            #TODO: auto broadcast_to support in numpy but not in our implemented NDArray
            return self.weight.broadcast_to(x.shape) * norm + self.bias.broadcast_to(x.shape)
        else:
            norm = (x - self.running_mean.broadcast_to(x.shape)) / ((self.running_var.broadcast_to(x.shape) + self.eps) ** 0.5)
            return self.weight.broadcast_to(x.shape) * norm + self.bias.broadcast_to(x.shape)

class LayerNorm1d(Module):
    def __init__(self, num_features: int, eps=1e-5, device=None, dtype=None):
        super().__init__()
        self.eps = eps
        
        self.weight = Parameter(init.ones(num_features))
        self.bias = Parameter(init.zeros(num_features))

    def forward(self, x: Tensor) -> Tensor:
        # x: (batch_size, C)
        mean = x.sum(dim=1, keepdim=True).broadcast_to(x.shape) / x.shape[1]
        var = ((x - mean)**2).sum(dim=1, keepdim=True).broadcast_to(x.shape) / x.shape[1]
        norm = (x - mean) / ((var + self.eps) ** 0.5)
        return self.weight.broadcast_to(x.shape) * norm + self.bias.broadcast_to(x.shape)

class Residual(Module):
    def __init__(self, fn: Module):
        super().__init__()
        self.fn = fn

    def forward(self, x: Tensor) -> Tensor:
        return x + self.fn(x)


import numpy as np

class Transform:
    def __call__(self, x):
        raise NotImplementedError


class RandomFlipHorizontal(Transform):
    def __init__(self, p = 0.5):
        self.p = p

    def __call__(self, img):
        """
        Horizonally flip an image, specified as an H x W x C NDArray.
        Args:
            img: H x W x C NDArray of an image
        Returns:
            H x W x C ndarray corresponding to image flipped with probability self.p
        Note: use the provided code to provide randomness, for easier testing
        """
        flip_img = np.random.rand() < self.p
        if flip_img:
            return img[:, ::-1, :]
            # return np.flip(img, axis=1)
        else:
            return img    
        


class RandomCrop(Transform):
    def __init__(self, padding=3):
        self.padding = padding

    def __call__(self, img):
        """ Zero pad and then randomly crop an image.
        Args:
             img: H x W x C NDArray of an image
        Return 
            H x W x C NAArray of cliped image
        Note: generate the image shifted by shift_x, shift_y specified below
        """
        shift_x, shift_y = np.random.randint(low=-self.padding, high=self.padding+1, size=2)
        padded_img = np.pad(img, ((self.padding, self.padding), (self.padding, self.padding), (0, 0)), 'constant')  
        H, W, _ = padded_img.shape
        return padded_img[self.padding + shift_x: H - self.padding + shift_x, self.padding + shift_y: W - self.padding + shift_y, :]



from ..data_basic import Dataset

class NDArrayDataset(Dataset):
    def __init__(self, *arrays):
        self.arrays = arrays

    def __len__(self) -> int:
        return self.arrays[0].shape[0]

    def __getitem__(self, i) -> object:
        return tuple([a[i] for a in self.arrays])


import flowlite as fl


def rand(*shape, low=0.0, high=1.0, device=None, dtype="float32", requires_grad=False):
    """Generate random numbers uniform between low and high"""
    device = fl.cpu() if device is None else device
    array = device.rand(*shape) * (high - low) + low
    return fl.Tensor(array, device=device, dtype=dtype, requires_grad=requires_grad)


def randn(*shape, mean=0.0, std=1.0, device=None, dtype="float32", requires_grad=False):
    """Generate random normal with specified mean and std deviation"""
    device = fl.cpu() if device is None else device
    array = device.randn(*shape) * std + mean
    return fl.Tensor(array, device=device, dtype=dtype, requires_grad=requires_grad)


def constant(*shape, c=1.0, device=None, dtype="float32", requires_grad=False):
    """Generate constant Tensor"""
    device = fl.cpu() if device is None else device
    array = device.ones(*shape, dtype=dtype) * c  # note: can change dtype
    return fl.Tensor(array, device=device, dtype=dtype, requires_grad=requires_grad)


def ones(*shape, device=None, dtype="float32", requires_grad=False):
    """Generate all-ones Tensor"""
    return constant(
        *shape, c=1.0, device=device, dtype=dtype, requires_grad=requires_grad
    )


def zeros(*shape, device=None, dtype="float32", requires_grad=False):
    """Generate all-zeros Tensor"""
    return constant(
        *shape, c=0.0, device=device, dtype=dtype, requires_grad=requires_grad
    )


def randb(*shape, p=0.5, device=None, dtype="bool", requires_grad=False):
    """Generate binary random Tensor"""
    device = fl.cpu() if device is None else device
    array = device.rand(*shape) <= p
    return fl.Tensor(array, device=device, dtype=dtype, requires_grad=requires_grad)


def one_hot(n, i, device=None, dtype="float32", requires_grad=False):
    """Generate one-hot encoding Tensor"""
    device = fl.cpu() if device is None else device
    return fl.Tensor(
        device.one_hot(n, i.numpy(), dtype=dtype),
        device=device,
        requires_grad=requires_grad,
    )


def zeros_like(array, *, device=None, requires_grad=False):
    device = device if device else array.device
    return zeros(
        *array.shape, dtype=array.dtype, device=device, requires_grad=requires_grad
    )


def ones_like(array, *, device=None, requires_grad=False):
    device = device if device else array.device
    return ones(
        *array.shape, dtype=array.dtype, device=device, requires_grad=requires_grad
    )



from ..autograd import Tensor
from typing import Optional, List
import numpy as np


class Dataset:
    r"""An abstract class representing a `Dataset`.

    All subclasses should overwrite :meth:`__getitem__`, supporting fetching a
    data sample for a given key. Subclasses must also overwrite
    :meth:`__len__`, which is expected to return the size of the dataset.
    """
    def __init__(self, transforms: Optional[List] = None):
        self.transforms = transforms
        
    def __getitem__(self, index) -> object:
        raise NotImplementedError()
    
    def apply_transforms(self, x):
        if self.transforms is not None:
            for trans in self.transforms:
                x = trans(x)
        return x
    
class DataLoader:
    r"""
    Data loader. Combines a dataset and a sampler, and provides an iterable over
    the given dataset.
    Args:
        dataset (Dataset): dataset from which to load the data.
        batch_size (int, optional): how many samples per batch to load
            (default: ``1``).
        shuffle (bool, optional): set to ``True`` to have the data reshuffled
            at every epoch (default: ``False``).
     """
    dataset: Dataset
    batch_size: Optional[int]

    def __init__(
        self,
        dataset: Dataset,
        batch_size: Optional[int] = 1,
        shuffle: bool = False,
    ):

        self.dataset = dataset
        self.shuffle = shuffle
        self.batch_size = batch_size
        if not self.shuffle:
            self.ordering = np.array_split(np.arange(len(dataset)), 
                                           range(batch_size, len(dataset), batch_size))

    def __iter__(self):
        if self.shuffle:
            self.ordering = np.array_split(np.random.permutation(len(self.dataset)), 
                                           range(self.batch_size, len(self.dataset), self.batch_size))
        self.cur_idx = 0
        return self

    def __next__(self):
        if self.cur_idx >= len(self.ordering):
            self.cur_idx = 0
            raise StopIteration
        batch_data = self.dataset[self.ordering[self.cur_idx]]
        batch_data = [Tensor(s) for s in batch_data]
        self.cur_idx += 1
        return batch_data

    